xquery version "3.0";

(:
 : Copyright 2006-2009 The FLWOR Foundation.
 :
 : Licensed under the Apache License, Version 2.0 (the "License");
 : you may not use this file except in compliance with the License.
 : You may obtain a copy of the License at
 :
 : http://www.apache.org/licenses/LICENSE-2.0
 :
 : Unless required by applicable law or agreed to in writing, software
 : distributed under the License is distributed on an "AS IS" BASIS,
 : WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 : See the License for the specific language governing permissions and
 : limitations under the License.
:)

module namespace bash = "http://www.zorba-xquery.com/modules/bash";

import module namespace file = "http://expath.org/ns/file";

import module namespace iomod = "http://www.zorba-xquery.com/modules/iomod";

import module namespace rand = "http://www.zorba-xquery.com/modules/random";

declare namespace an = "http://www.zorba-xquery.com/annotations";

declare namespace ver = "http://www.zorba-xquery.com/options/versioning";
declare option ver:module-version "1.0";

declare variable $bash:stdin := xs:anyURI('http://www.zorba-xquery.com/modules/bash/file/stdin');
declare variable $bash:stdout := xs:anyURI('http://www.zorba-xquery.com/modules/bash/file/stdout');
declare variable $bash:stderr := xs:anyURI('http://www.zorba-xquery.com/modules/bash/file/stderr');
declare variable $bash:pipe-in := xs:anyURI('http://www.zorba-xquery.com/modules/bash/pipe/in');
declare variable $bash:pipe-out := xs:anyURI('http://www.zorba-xquery.com/modules/bash/pipe/out');
declare variable $bash:null := xs:anyURI('http://www.zorba-xquery.com/modules/bash/file/null');

declare variable $bash:default-fd-mapping := ($bash:stdin, $bash:stdout, $bash:stderr, for $i in 3 to 9 return $bash:null);
declare variable $bash:standard-fd-mapping := ($bash:stdin, $bash:pipe-out, $bash:stderr, for $i in 3 to 9 return $bash:null);

(: Launches an exec command with the supplied command-line parameters, and with the supplied file descriptor mapping :)
(: A file descriptor set to bash:stdin can be read, and the input comes from stdin :)
(: A file descriptor set to bash:stdout can be written to, and the output is sent to stdout :)
(: A file descriptor set to bash:stderr can be written to, and the output is sent to stderr :)
(: A file descriptor set to bash:null cannot be read or written :)
(: A file descriptor set to bash:pipe-in can be read, and the input is taken from the third parameter
  (each xs:string item is used for an input pipe, in the same order) :)
(: A file descriptor set to bash:pipe-out can be written to, and the output produced is returned by the function (one item for each output pipe, in the file descriptor order) :)
declare %an:sequential function bash:exec($args as xs:string*, $pipe-ins as xs:int*, 
  $input-streams as xs:string*, $pipe-outs as xs:int*) as xs:string* external;

declare %an:sequential function bash:exec($args as xs:string*, $pipe-ins as xs:int*, 
  $input-streams as xs:string*) as xs:string* {
  bash:exec($args, $pipe-ins, $input-streams, ())
};

(: Executes exec with the supplied parameters, with the supplied file descriptor mapping, piping in being disallowed :)
declare %an:sequential function bash:exec($args as xs:string*, $pipe-outs as xs:int*) as xs:string* {
  bash:exec($args, (), (), $pipe-outs)
};

(: Executes exec with the supplied parameters, with the default file descriptor mapping :)
declare %an:sequential function bash:exec($args as xs:string*) as empty-sequence() {
  bash:exec($args, (), (), ())
};

(: Get the current directory path :)
declare %an:nondeterministic function bash:pwd() as xs:string external; 

(: Get the items in current directory:)
declare %an:nondeterministic function bash:ls() as xs:string* {
  file:list(bash:pwd())
}; 

(: Get lines matching a pattern :)
declare function bash:grep($pattern as xs:string, $input as xs:string*) as xs:string* {
  for $line in $input
  return
  if (fn:matches($line, $pattern)) then
    $line
  else
    ()
}; 

(: Count the number of lines, words or characters in the input stream :)
declare %an:sequential function bash:wc($input_stream as xs:string, $params as xs:string*) as xs:integer* {
  for $p in $params
  return
  if ($p = "l") then
    fn:count(iomod:readlines($input_stream))
  else if ($p = "w") then
    fn:count(iomod:readwords($input_stream))
  else if ($p = "c") then
    fn:count(iomod:readchars($input_stream))
  else
    ()
}; 

(: sed command in linux, currently only a wrapper of replace TODO- replace will materialize the streamable string :)
declare function bash:sed($input_stream as xs:string, 
                         $pattern as xs:string,
                         $replacement as xs:string) as xs:string {
  fn:replace($input_stream, $pattern, $replacement)
}; 

(: sed command in linux, currently only a wrapper of replace :)
declare function bash:sed($input_stream as xs:string, 
                         $pattern as xs:string,
                         $replacement as xs:string,
                         $flags as xs:string) as xs:string {
  fn:replace($input_stream, $pattern, $replacement, $flags)
}; 

(: Sort lines of an input stream TODO- external sort to process big data :)
declare %an:sequential function bash:sort($input_stream as xs:string) as xs:string* {
  let $input_sequence := iomod:readlines($input_stream)
  for $line in $input_sequence
  order by $line
  return $line
}; 

(: Sort lines of an input stream in reverse order :)
declare %an:sequential function bash:rsort($input_stream as xs:string) as xs:string* {
  let $input_sequence := iomod:readlines($input_stream)
  for $line in $input_sequence
  order by $line descending
  return $line
}; 

(: Return lines beginning with a given string :)
declare %an:sequential function bash:look($input_stream as xs:string, $str as xs:string) as xs:string* {
  for $line in iomod:readlines($input_stream)
  where fn:starts-with($line, $str)
  return $line
}; 

(: Implementation of shuffle, stackoverflow? :)
declare %private %an:nondeterministic function bash:shuffle_impl($input, $result) {
  let $len := fn:count($input)
  let $index := if ($len = 0) then () else rand:random-between(1, fn:count($input))
  let $new_result := if ($len = 0) then () else fn:insert-before($result, 1, $input[$index])
  let $new_input := if ($len = 0) then () else fn:remove($input, $index)
  return
  if ($len = 0) then
    $result
  else 
    bash:shuffle_impl($new_input, $new_result)
};

(: Return the input sequence in a randome order :)
declare %an:nondeterministic function bash:shuffle($input_sequence) {
  bash:shuffle_impl($input_sequence, ())
}; 

(: Print a string to stdout until interrupt :)
declare %an:sequential function bash:yes($str) {
  while (true()) {
    iomod:writeline($iomod:stdout, $str);
  }
}; 

declare %an:sequential function bash:yes() {
  while (true()) {
    iomod:writeline($iomod:stdout, "y");
  }
};

