/*
 * Copyright 2006-2008 The FLWOR Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <iostream>
#include <vector>
#include <sstream>
#include <string>
#include <cstring>
#include <cassert>
#include <cstdio>

#include <unistd.h>
#include <errno.h>

#include <zorba/zorba.h>
#include <zorba/external_module.h>
#include <zorba/user_exception.h>
#include <zorba/function.h>
#include <zorba/empty_sequence.h>
#include <zorba/item_sequence.h>
#include <zorba/vector_item_sequence.h>
#include <zorba/xquery_exception.h>
#include <zorba/singleton_item_sequence.h>
#include <zorba/item_factory.h>

#include "FileBuf.h"

using namespace zorba;

namespace zorba { namespace bashmodule {
  
class ExecExternalFunction;
class PwdExternalFunction;

class BashExternalModule : public ExternalModule
{
private:
  static ItemFactory *theFactory; 

protected:
  ExecExternalFunction *func_exec;
  PwdExternalFunction *func_pwd;

public:
  BashExternalModule()
    :
    func_exec(0),
    func_pwd(0)
  {
  }

  virtual ~BashExternalModule();

  virtual String getURI() const { return "http://www.zorba-xquery.com/modules/bash"; }

  virtual ExternalFunction* getExternalFunction(const String& aLocalname);
  
  virtual void destroy();

  static String bash_stdin;
  static String bash_stdout;
  static String bash_stderr;
  static String bash_pipein;
  static String bash_pipeout;
  static String bash_null;

  static void streamReleaser(std::istream* stream)
  {
    FileBuf *fbuf = dynamic_cast<FileBuf*>(stream->rdbuf());
    FILE *fd = fbuf->getFILE();
    fclose(fd);
    delete fbuf;
    delete stream;
  }

  static ItemFactory* getItemFactory()
  {
    if (!theFactory)
    {
      theFactory = Zorba::getInstance(0)->getItemFactory();
    }
    return theFactory;
  }
};

#define READ  0
#define WRITE 1

pid_t zorba_popen(const char *command, std::istream *istr, int *outfp, int *errfp)
{
    int p_stdin[2];
    int p_stdout[2];
    int p_stderr[2];
    pid_t pid;
    pid_t driver_pid;

    if (pipe(p_stdin) != 0 || pipe(p_stdout) != 0 || pipe(p_stderr) != 0)
    {
      std::stringstream lErrorMsg;
      lErrorMsg << "Failed to create pipe";
      Item lQName = BashExternalModule::getItemFactory()->createQName(
      "http://www.zorba-xquery.com/modules/bash", "PROC01");
      throw USER_EXCEPTION(lQName, lErrorMsg.str().c_str());
    }

    pid = fork();

    if (pid < 0)
    {
      return pid;
    }
    else if (pid == 0)
    {
      close(p_stdin[WRITE]);
      dup2(p_stdin[READ], 0);   // duplicate stdin

      close(p_stdout[READ]);
      dup2(p_stdout[WRITE], 1); // duplicate stdout

      close(p_stderr[READ]);
      dup2(p_stderr[WRITE], 2); // duplicate stderr

      FILE *f_out = fdopen(p_stdout[WRITE], "w");
      if (f_out == NULL)
      {
        std::stringstream lErrorMsg;
        lErrorMsg << "Failed to call fdopen";
        Item lQName = BashExternalModule::getItemFactory()->createQName(
        "http://www.zorba-xquery.com/modules/bash", "PROC01");
        throw USER_EXCEPTION(lQName, lErrorMsg.str().c_str());
      }

      execl("/bin/sh", "sh", "-c", command, NULL);

      // Only return when error ocurrs
      perror("execl"); // output the result to standard error
      exit(errno);
    }
    
    if (istr != NULL)
    {
      driver_pid = fork();
      if (driver_pid < 0)
      {
        std::stringstream lErrorMsg;
        lErrorMsg << "Failed to execute the command fork driver process";
        Item lQName = BashExternalModule::getItemFactory()->createQName(
        "http://www.zorba-xquery.com/modules/bash", "PROC01");
        throw USER_EXCEPTION(lQName, lErrorMsg.str().c_str());
        exit(-1);
      }
      else if (driver_pid == 0)
      {
        // This process write the stream to the process created above
        close(p_stdin[READ]);
        dup2(p_stdin[WRITE], 1);
        
        close(p_stdout[READ]);
        close(p_stdout[WRITE]);
        close(p_stderr[READ]);
        close(p_stderr[WRITE]);

        char c;
        while (istr->good())
        {
          c = istr->get();
          std::cout << c;
        }
        exit(0);
      }
    }

    if (outfp == NULL)
    {
      close(p_stdout[READ]);
    }
    else
    {
      *outfp = p_stdout[READ];
    }

    if (errfp == NULL)
    {
      close(p_stderr[READ]);
    }
    else
    {
      *errfp = p_stderr[READ];
    }

    close(p_stdin[WRITE]);
    close(p_stdin[READ]);
    close(p_stdout[WRITE]);
    close(p_stderr[WRITE]);
        
    return pid;
}

class ExecExternalFunction : public ContextualExternalFunction
{
protected:
  const BashExternalModule *theModule;

public:
  ExecExternalFunction(const BashExternalModule *aModule)
    :
    theModule(aModule)
  {
  }

  String getURI() const { return theModule->getURI(); }

  String getLocalName() const { return "exec"; }

  ItemSequence_t evaluate(
      const ExternalFunction::Arguments_t &args,
      const StaticContext *sctx,
      const DynamicContext *dctx) const
  {
    StringVector command_str;
    StringVector mapping;
    std::vector<Item> targets;
    
    // 1. Get the command line
    Item item;
    Iterator_t iter = args[0]->getIterator();
    iter->open();
    while (iter->next(item))
    {
      command_str.push_back(item.getStringValue());
    }
    iter->close();
    // concat the command strings
    String command;
    for (size_t i = 0; i < command_str.size(); i++)
    {
      command += command_str[i] + " ";
    }
    
    // 2. Get the mapping of file descriptors
    iter = args[1]->getIterator();
    iter->open();
    while (iter->next(item))
    {
      mapping.push_back(item.getStringValue());
    }
    iter->close();

    // 3. Get the targets of redirection, if any
    iter = args[2]->getIterator();
    iter->open();
    while (iter->next(item))
    {
      targets.push_back(item);
    }
    iter->close();

    std::vector<int> pipeout;
    for (size_t i = 0; i < mapping.size(); i++)
    {
      if (mapping[i].compare(BashExternalModule::bash_pipeout) == 0)
      {
        pipeout.push_back(i);
      }
    }

    std::vector<Item> results;
    int outfp;
    int errfp;
    int *outptr = NULL;
    int *errptr = NULL;
    std::istream *istr = NULL;
    pid_t pid;

    // Pipe the standard out or error
    for (size_t i = 0; i < pipeout.size(); i++)
    {
      if (pipeout[i] == 1)
      {
        outptr = &outfp;
      }
      if (pipeout[i] == 2)
      {
        errptr = &errfp;
      }
    }

    // Pipe the standard in from an input stream
    int pipein = 0;
    if (mapping[0].compare(BashExternalModule::bash_pipein) == 0)
    {
      istr = &(targets[pipein++].getStream());
    }
    pid = zorba_popen(command.str().c_str(), istr, outptr, errptr);
    if (pid == -1)
    {
      std::stringstream lErrorMsg;
      lErrorMsg << "Failed to execute the command (" << command << ")";
      Item lQName = BashExternalModule::getItemFactory()->createQName(
      "http://www.zorba-xquery.com/modules/bash", "PROC01");
      throw USER_EXCEPTION(lQName, lErrorMsg.str().c_str());
    }
    else
    {
      if (outptr != NULL)
      {
        FILE *fptr = fdopen(*outptr, "r");
        FileBuf *fbuf = new FileBuf(fptr);
        std::istream *is = new std::istream(fbuf);
        Item stritem = BashExternalModule::getItemFactory()->createStreamableString(*is, &BashExternalModule::streamReleaser, true);
        results.push_back(stritem);
      }
      if (errptr != NULL)
      {
        FILE *fptr = fdopen(*errptr, "r");
        FileBuf *fbuf = new FileBuf(fptr);
        std::istream *is = new std::istream(fbuf);
        results.push_back(BashExternalModule::getItemFactory()->createStreamableString(*is, &BashExternalModule::streamReleaser, true));
      }
    }

    if (results.size() == 0)
    {
      return ItemSequence_t(new EmptySequence());
    }
    else
    {
      return ItemSequence_t(new VectorItemSequence(results));
    }
  }
  
private:
  typedef std::vector<String> StringVector;
};

class PwdExternalFunction : public ContextualExternalFunction
{
protected:
  const BashExternalModule *theModule;

public:
  PwdExternalFunction(const BashExternalModule *aModule)
    :
    theModule(aModule)
  {
  }

  String getURI() const { return theModule->getURI(); }

  String getLocalName() const { return "pwd"; }

  ItemSequence_t evaluate(
      const ExternalFunction::Arguments_t &args,
      const StaticContext *sctx,
      const DynamicContext *dctx) const
  {
    char buffer[2048];
    getcwd(buffer, 2048);
    String result(buffer);
    Item item = BashExternalModule::getItemFactory()->createString(result);
    return ItemSequence_t(new SingletonItemSequence(item));
  }
  
};

ItemFactory* BashExternalModule::theFactory = 0;

String BashExternalModule::bash_stdin = "http://www.zorba-xquery.com/modules/bash/file/stdin";
String BashExternalModule::bash_stdout = "http://www.zorba-xquery.com/modules/bash/file/stdout";
String BashExternalModule::bash_stderr = "http://www.zorba-xquery.com/modules/bash/file/stderr";
String BashExternalModule::bash_pipein = "http://www.zorba-xquery.com/modules/bash/pipe/in";
String BashExternalModule::bash_pipeout = "http://www.zorba-xquery.com/modules/bash/pipe/out";
String BashExternalModule::bash_null = "http://www.zorba-xquery.com/modules/bash/file/null";

BashExternalModule::~BashExternalModule()
{
  delete func_exec;
  delete func_pwd;
}

void
BashExternalModule::destroy()
{
  if (!dynamic_cast<BashExternalModule*>(this)) {
    return;
  }
  delete this;
}

ExternalFunction* BashExternalModule::getExternalFunction(const String& aLocalName)
{
  if (aLocalName == "exec")
  {
    if (!func_exec)
    {
      func_exec = new ExecExternalFunction(this);
    }
    return func_exec;
  }
  if (aLocalName == "pwd")
  {
    if (!func_pwd)
    {
      func_pwd = new PwdExternalFunction(this);
    }
    return func_pwd;
  }
  return 0;
}

}}

#ifdef WIN32
#  define DLL_EXPORT __declspec(dllexport)
#else
#  define DLL_EXPORT __attribute__ ((visibility("default")))
#endif

extern "C" DLL_EXPORT zorba::ExternalModule* createModule() {
  return new zorba::bashmodule::BashExternalModule();
}

